#include <ntddk.h>
#include <common.h>
#include <wr_private.h>

WCHAR DeviceName[] = L"\\Device\\nodel";
WCHAR SymLinkName[] = L"\\DosDevices\\nodel";

UNICODE_STRING usDeviceName;
UNICODE_STRING usSymbolicLinkName;

typedef struct _DEVICE_CONTEXT
{
	PDRIVER_OBJECT  pDriverObject;       
    PDEVICE_OBJECT  pDeviceObject;
}
DEVICE_CONTEXT, *PDEVICE_CONTEXT, **PPDEVICE_CONTEXT;

PDEVICE_OBJECT  g_pDeviceObject  = NULL;
PDEVICE_CONTEXT g_pDeviceContext = NULL;

#define FILE_DEVICE_NODEL 0x8000

#define CODE_NODEL_INFO	CTL_CODE(FILE_DEVICE_NODEL, 0x800, \
	METHOD_BUFFERED, FILE_ANY_ACCESS)


NTSTATUS DriverInitialize(PDRIVER_OBJECT  pDriverObject,
						  PUNICODE_STRING pusRegistryPath);

NTSTATUS DriverEntry(PDRIVER_OBJECT  pDriverObject,
					 PUNICODE_STRING pusRegistryPath);

#ifdef ALLOC_PRAGMA

#pragma alloc_text (INIT, DriverInitialize)
#pragma alloc_text (INIT, DriverEntry)

#endif

//
// undeclared functions
//

NTSYSAPI NTSTATUS NTAPI ObQueryNameString(POBJECT Object, 
										  PUNICODE_STRING Name, 
										  ULONG MaximumLength, 
										  PULONG ActualLength);

//
// SDT hook stuff
// 

#define MAX_PATH 260
WCHAR PathName[0x1000] = { 0 };
UNICODE_STRING usPathName;

extern PSERVICE_DESCRIPTOR_TABLE KeServiceDescriptorTable;

PVOID *KeServiceTablePointers;
PMDL KeServiceTableMdl;
BOOLEAN *ServiceIsHooked;

PPVOID MapServiceTable(BOOLEAN **);
VOID UnmapServiceTable(PVOID);

#ifdef ALPHA
#define FUNCTION_PTR(_Function) (*(PULONG) _Function) & 0x0000FFFF
#else
#define FUNCTION_PTR(_Function) *(PULONG)((PUCHAR) _Function + 1)
#endif

#define HOOK_SYSCALL(_Function, _Hook, _Orig)							\
	if (!ServiceIsHooked[FUNCTION_PTR(_Function)]) {					\
	_Orig = (PVOID) InterlockedExchange((PLONG)							\
	&KeServiceTablePointers[FUNCTION_PTR(_Function)], (LONG) _Hook );	\
	ServiceIsHooked[ FUNCTION_PTR(_Function) ] = TRUE; }


#define UNHOOK_SYSCALL(_Function, _Hook, _Orig)								\
	if (ServiceIsHooked[FUNCTION_PTR(_Function)] &&							\
	KeServiceTablePointers[FUNCTION_PTR(_Function) ] == (PVOID) _Hook ) {	\
	InterlockedExchange((PLONG) &KeServiceTablePointers[					\
	FUNCTION_PTR(_Function)], (LONG) _Orig );								\
	ServiceIsHooked[FUNCTION_PTR(_Function)] = FALSE; }


NTSTATUS (*RealZwSetInformationFile)(IN HANDLE FileHandle,
									 OUT PIO_STATUS_BLOCK IoStatusBlock,
									 IN PVOID FileInformation,
									 IN ULONG Length,
									 IN FILE_INFORMATION_CLASS FileInformationClass);

NTSTATUS HookedZwSetInformationFile(IN HANDLE FileHandle,
									OUT PIO_STATUS_BLOCK IoStatusBlock,
									IN PVOID FileInformation,
									IN ULONG Length,
									IN FILE_INFORMATION_CLASS FileInformationClass)
{
	PUNICODE_STRING us;
	POBJECT pObj;
	ULONG Len;

	FILE_DISPOSITION_INFORMATION *pDisposition;

	if (PathName[0] == 0)
		goto ret;

	if (FileInformationClass == FileDispositionInformation)
	{
		pDisposition = (FILE_DISPOSITION_INFORMATION *) FileInformation;

		if (pDisposition->DeleteFile == TRUE)
		{
			us = malloc(0x1000);

			if (us)
			{
				memset(us, 0, 0x1000);

				//
				// it's not right but who cares, 
				// this is much more than the normal size
				//

				us->Length = us->MaximumLength = 0x1000;

				if (ObReferenceObjectByHandle(FileHandle, 0, NULL, 
					KernelMode, &pObj, NULL) == STATUS_SUCCESS)
				{
					//
					// we can user ObQueryNameString since our
					// driver is not a fs filter
					//

					if (ObQueryNameString(pObj, us, 
						us->Length, &Len) == STATUS_SUCCESS) 
					{
						if (us->Length >= usPathName.Length)
						{
							//
							// i'm very lazy so to compare the strings
							// i just truncate the file string
							// to compare it with the path string
							//

							us->Buffer[usPathName.Length / 2] = 0;
							us->Length = usPathName.Length;
							us->MaximumLength = usPathName.MaximumLength;

							if (RtlCompareUnicodeString(&usPathName, us, TRUE) == 0)
							{
								DbgPrint("Protected file\n");

								return STATUS_FILE_INVALID;
							}
						}
					}

					ObDereferenceObject(pObj);
				}

				free(us);
			}
		}
	}

ret:

	return RealZwSetInformationFile(FileHandle, IoStatusBlock, 
		FileInformation, Length, FileInformationClass);
}


PPVOID MapServiceTable(BOOLEAN **ServiceIsHooked)
{
	PVOID Mem;

	Mem = ExAllocatePoolWithTag(0, 
		KeServiceDescriptorTable->ntoskrnl.ServiceLimit,
		0x206B6444);

	if (Mem == NULL)
		return NULL;

	*ServiceIsHooked = (BOOLEAN *) Mem;

	memset(Mem, 0, KeServiceDescriptorTable->ntoskrnl.
		ServiceLimit);

	KeServiceTableMdl = MmCreateMdl(NULL, 
		KeServiceDescriptorTable->ntoskrnl.ServiceTable,
		(KeServiceDescriptorTable->ntoskrnl.ServiceLimit * 
		sizeof (POINTER)));

	if (KeServiceTableMdl == NULL)
		return NULL;

	MmBuildMdlForNonPagedPool(KeServiceTableMdl);

	return (PPVOID) MmMapLockedPages(KeServiceTableMdl, 0);
}

VOID UnmapServiceTable(PVOID KeServiceTablePointers)
{
	if (KeServiceTableMdl == NULL)
		return;

	MmUnmapLockedPages(KeServiceTablePointers,
		KeServiceTableMdl);

	ExFreePool(KeServiceTableMdl);
}

//
// patch KeAttachProcess
//
/*
VOID RebuildNtoskrnl()
{
	PMDL Mdl;
	DWORD CR0Backup;
	DWORD *ptr;
	BYTE *Patch;

	ptr = (DWORD *)(2 + (DWORD) &KeAttachProcess);

	ptr = (DWORD*) *ptr;

	Patch = (BYTE *) *ptr;

	Mdl = MmCreateMdl(0, (PVOID) Patch, 5);

	if (Mdl == NULL)
		return;

	MmProbeAndLockPages(Mdl, 0, 0);

	if (*Patch == 0xE9)
	{
		__asm
		{
			mov eax, cr0
			mov CR0Backup, eax
			and eax, 0xFFFEFFFF
			mov cr0, eax
		}

		//
		// patch
		//

		memcpy(Patch, RestoreInfo.KeAttachProcessPatch, 5);

		__asm
		{
			mov eax, CR0Backup
			mov cr0, eax
		}
	}

	MmUnlockPages(Mdl);
	IoFreeMdl(Mdl);
}
*/

NTSTATUS ControlDispatcher(PDEVICE_CONTEXT pDeviceContext, DWORD dwCode,
						   BYTE *pInput, DWORD dwInputSize,
						   BYTE *pOutput, DWORD dwOutputSize, DWORD *pdwInfo)
{
	switch (dwCode)
	{

	case CODE_NODEL_INFO:
		{
			memcpy(PathName, pInput, 0x1000 * sizeof (WCHAR));

			DbgPrint("%ws\n", PathName);

			RtlInitUnicodeString(&usPathName, PathName);

			break;
		}
		
	default:
	 	return STATUS_INVALID_PARAMETER;
	}
	 
	return STATUS_SUCCESS;
}

NTSTATUS DeviceDispatcher(PDEVICE_CONTEXT pDeviceContext, PIRP pIrp)
{
    PIO_STACK_LOCATION pisl;
    DWORD dwInfo = 0;
    NTSTATUS ns = STATUS_NOT_IMPLEMENTED;

    pisl = IoGetCurrentIrpStackLocation(pIrp);

    switch (pisl->MajorFunction)
	{
	
	case IRP_MJ_CREATE:
	case IRP_MJ_CLEANUP:
	case IRP_MJ_CLOSE:
		{
			ns = STATUS_SUCCESS;
            break;
		}

	case IRP_MJ_DEVICE_CONTROL:
		{

			ns = ControlDispatcher(pDeviceContext,
				pisl->Parameters.DeviceIoControl.IoControlCode,
				(BYTE *) pIrp->AssociatedIrp.SystemBuffer,
				pisl->Parameters.DeviceIoControl.InputBufferLength,
				(BYTE *) pIrp->AssociatedIrp.SystemBuffer,
				pisl->Parameters.DeviceIoControl.OutputBufferLength,
				&dwInfo);
			
			break;
		}
	}

    pIrp->IoStatus.Status = ns;
    pIrp->IoStatus.Information = dwInfo;

    IoCompleteRequest (pIrp, IO_NO_INCREMENT);

    return ns;
}

NTSTATUS DriverDispatcher(PDEVICE_OBJECT pDeviceObject, PIRP pIrp)
{
	return (pDeviceObject == g_pDeviceObject ? 
		DeviceDispatcher(g_pDeviceContext, pIrp)
		: STATUS_INVALID_PARAMETER_1);
}


VOID DriverUnload(PDRIVER_OBJECT pDriverObject)
{
	UNHOOK_SYSCALL(ZwSetInformationFile, 
		HookedZwSetInformationFile, RealZwSetInformationFile);
	UnmapServiceTable(KeServiceTablePointers);

    IoDeleteSymbolicLink(&usSymbolicLinkName);
    IoDeleteDevice(pDriverObject->DeviceObject);
}

NTSTATUS DriverInitialize(PDRIVER_OBJECT pDriverObject,
						  PUNICODE_STRING pusRegistryPath)
{
    PDEVICE_OBJECT pDeviceObject = NULL;
    NTSTATUS ns = STATUS_DEVICE_CONFIGURATION_ERROR;
    
    RtlInitUnicodeString(&usDeviceName, DeviceName);
    RtlInitUnicodeString(&usSymbolicLinkName, SymLinkName);

    if ((ns = IoCreateDevice(pDriverObject, sizeof (DEVICE_CONTEXT),
		&usDeviceName, FILE_DEVICE_NODEL, 0, FALSE,
		&pDeviceObject)) == STATUS_SUCCESS)
	{
        if ((ns = IoCreateSymbolicLink(&usSymbolicLinkName, 
			&usDeviceName)) == STATUS_SUCCESS)
		{
            g_pDeviceObject  = pDeviceObject;
            g_pDeviceContext = pDeviceObject->DeviceExtension;

            g_pDeviceContext->pDriverObject = pDriverObject;
            g_pDeviceContext->pDeviceObject = pDeviceObject;

		}
		else
		{
			IoDeleteDevice(pDeviceObject);
		}
	}
	
	return ns;
}

NTSTATUS DriverEntry(PDRIVER_OBJECT pDriverObject,
					 PUNICODE_STRING pusRegistryPath)
{
	PDRIVER_DISPATCH *ppdd;
    NTSTATUS ns = STATUS_DEVICE_CONFIGURATION_ERROR;

    if ((ns = DriverInitialize(pDriverObject, pusRegistryPath)) == STATUS_SUCCESS)
	{
        ppdd = pDriverObject->MajorFunction;

        ppdd[IRP_MJ_CREATE                  ] =
        ppdd[IRP_MJ_CREATE_NAMED_PIPE       ] =
        ppdd[IRP_MJ_CLOSE                   ] =
        ppdd[IRP_MJ_READ                    ] =
        ppdd[IRP_MJ_WRITE                   ] =
        ppdd[IRP_MJ_QUERY_INFORMATION       ] =
        ppdd[IRP_MJ_SET_INFORMATION         ] =
        ppdd[IRP_MJ_QUERY_EA                ] =
        ppdd[IRP_MJ_SET_EA                  ] =
        ppdd[IRP_MJ_FLUSH_BUFFERS           ] =
        ppdd[IRP_MJ_QUERY_VOLUME_INFORMATION] =
        ppdd[IRP_MJ_SET_VOLUME_INFORMATION  ] =
        ppdd[IRP_MJ_DIRECTORY_CONTROL       ] =
        ppdd[IRP_MJ_FILE_SYSTEM_CONTROL     ] =
        ppdd[IRP_MJ_DEVICE_CONTROL          ] =
        ppdd[IRP_MJ_INTERNAL_DEVICE_CONTROL ] =
        ppdd[IRP_MJ_SHUTDOWN                ] =
        ppdd[IRP_MJ_LOCK_CONTROL            ] =
        ppdd[IRP_MJ_CLEANUP                 ] =
        ppdd[IRP_MJ_CREATE_MAILSLOT         ] =
        ppdd[IRP_MJ_QUERY_SECURITY          ] =
        ppdd[IRP_MJ_SET_SECURITY            ] =
        ppdd[IRP_MJ_POWER                   ] =
        ppdd[IRP_MJ_SYSTEM_CONTROL          ] =
        ppdd[IRP_MJ_DEVICE_CHANGE           ] =
        ppdd[IRP_MJ_QUERY_QUOTA             ] =
        ppdd[IRP_MJ_SET_QUOTA               ] =
        ppdd[IRP_MJ_PNP                     ] = DriverDispatcher;
        pDriverObject->DriverUnload           = DriverUnload;

		KeServiceTablePointers = MapServiceTable(&ServiceIsHooked);

		HOOK_SYSCALL(ZwSetInformationFile, 
			HookedZwSetInformationFile, RealZwSetInformationFile);

		//DbgPrint("%08X - %08X\n", ZwSetInformationFile, 
		//	HookedZwSetInformationFile);

	}

	return ns;
}
