#include <ntddk.h>
#include <common.h>
#include <wr_private.h>

HANDLE WrCreateFileW(IN PWCHAR pName, IN ACCESS_MASK  DesiredAccess,
					 IN ULONG FileAttributes, IN ULONG ShareAccess,
					 IN ULONG CreateDisposition)
{
	NTSTATUS			nt;
	UNICODE_STRING		usFileName;
    OBJECT_ATTRIBUTES	ObjAttr;
    IO_STATUS_BLOCK		IoStatus;
	HANDLE				hFile;

	RtlInitUnicodeString(&usFileName, pName);

    InitializeObjectAttributes(&ObjAttr, &usFileName,
		OBJ_KERNEL_HANDLE, NULL, NULL);

	nt = ZwCreateFile(&hFile, DesiredAccess,
		&ObjAttr, &IoStatus, NULL, 
		FileAttributes, ShareAccess,
		CreateDisposition, FILE_SYNCHRONOUS_IO_NONALERT, 
		NULL, 0);

	if (nt != STATUS_SUCCESS)
		return INVALID_HANDLE_VALUE;

	return hFile;
}

HANDLE WrCreateFileA(IN PCHAR pName, IN ACCESS_MASK  DesiredAccess,
					 IN ULONG FileAttributes, IN ULONG ShareAccess,
					 IN ULONG CreateDisposition)
{
	NTSTATUS			nt;
	ANSI_STRING			asFileName;
	UNICODE_STRING		usFileName;
    OBJECT_ATTRIBUTES	ObjAttr;
    IO_STATUS_BLOCK		IoStatus;
	HANDLE				hFile;

	RtlInitAnsiString(&asFileName, pName);

	RtlAnsiStringToUnicodeString(&usFileName,
		&asFileName, TRUE);

    InitializeObjectAttributes(&ObjAttr, &usFileName,
		OBJ_CASE_INSENSITIVE | OBJ_KERNEL_HANDLE, NULL, NULL);

	nt = ZwCreateFile(&hFile, DesiredAccess,
		&ObjAttr, &IoStatus, NULL, 
		FileAttributes, ShareAccess,
		CreateDisposition, FILE_SYNCHRONOUS_IO_NONALERT, 
		NULL, 0);

	RtlFreeUnicodeString(&usFileName);


	return nt == STATUS_SUCCESS ? hFile : INVALID_HANDLE_VALUE;
}

VOID WrCloseHandle(IN HANDLE hObj)
{
	ZwClose(hObj);
}

BOOL WrWriteFile(IN HANDLE hFile, IN PVOID pBuffer, 
				 IN ULONG uSize)
{
	NTSTATUS		nt;
	IO_STATUS_BLOCK IoStatus;
	LONGLONG		FileSize;

	nt = ZwWriteFile(hFile, NULL, NULL, NULL, &IoStatus,
		pBuffer, uSize, NULL, NULL);

	if (WrGetFileSize(hFile, &FileSize))
	{
		WrSetFilePointer(hFile, FileSize);
	}
	
	return nt == STATUS_SUCCESS ? TRUE : FALSE;
}

BOOL WrReadFile(IN HANDLE hFile, IN PVOID pBuffer, 
				IN ULONG uSize)
{
	NTSTATUS		nt;
	IO_STATUS_BLOCK IoStatus;

	nt = ZwReadFile(hFile, NULL, NULL, NULL, &IoStatus,
		pBuffer, uSize, NULL, NULL);
	
	return nt == STATUS_SUCCESS ? TRUE : FALSE;
}

BOOL WrSetFilePointer(IN HANDLE hFile, IN LONGLONG loloFilePtr)
{
	NTSTATUS nt;
	IO_STATUS_BLOCK IoStatus;
	FILE_POSITION_INFORMATION FilePosInfo;

	FilePosInfo.CurrentByteOffset.QuadPart = loloFilePtr;

	nt = ZwSetInformationFile(hFile, &IoStatus, &FilePosInfo, 
		sizeof (FILE_POSITION_INFORMATION), FilePositionInformation);

	return nt == STATUS_SUCCESS ? TRUE : FALSE;
}

BOOL WrGetFileSize(IN HANDLE hFile, IN PLONGLONG ploloFileSize)
{
	NTSTATUS Status;
    IO_STATUS_BLOCK IoStatus;
    FILE_STANDARD_INFORMATION StandardInfo;

    Status = ZwQueryInformationFile(hFile, &IoStatus,
                &StandardInfo, sizeof (StandardInfo),
                FileStandardInformation);

    if (!NT_SUCCESS(Status)) 
	{
        return FALSE;
	}
    else 
	{
        *ploloFileSize = StandardInfo.EndOfFile.QuadPart;
		return TRUE;
	}
}

